(ns com.aurellem.gb-funs
    (:import com.aurellem.gb.Gb)
    (:import java.io.File)
    (:import org.apache.commons.io.FileUtils)
    (:import (java.nio IntBuffer ByteOrder)))

;; Savestates
(defrecord SaveState [frame data])

(def ^:dynamic *save-state-cache*
  (File. "/home/r/proj/pokemon-escape/save-states/"))

(defn frame->filename [frame]
  (File. *save-state-cache* (format "%07d.sav" frame)))
  
(defn write-save! [^SaveState save]
  (let [buf (:data save)
        bytes (byte-array (.limit buf))
        dest (frame->filename (:frame save))]
    (.get buf bytes)
    (FileUtils/writeByteArrayToFile dest bytes)
    (.rewind buf)
    save))

(defn read-save [frame]
  (let [save (frame->filename frame)]
    (if (.exists save)
      (let [buf (Gb/saveBuffer)
            bytes (FileUtils/readFileToByteArray save)]
        (.put buf bytes)
        (.flip buf)
        (SaveState. frame buf)))))
;;;;;;;;;;;;;;;;

;; Gameboy management
(Gb/loadVBA)

(def yellow-rom-image
  (File. "/home/r/proj/pokemon-escape/roms/yellow.gbc"))

(def yellow-save-file
  (File. "/home/r/proj/pokemon-escape/roms/yellow.sav"))

(def on? (atom nil))

(defn shutdown! [] (Gb/shutdown) (reset! on? false))

(defn restart! []
  (shutdown!)
  (.delete yellow-save-file)
  (Gb/startEmulator (.getCanonicalPath yellow-rom-image))
  (reset! on? true))

;;; The first state!
(defn gen-root! []
  (restart!)
  (write-save! (SaveState. 0  (Gb/saveState))))

(defn root []
  (if (.exists (frame->filename 0))
    (read-save 0)
    (gen-root!)))

;;;; Press Buttons

(def button-code
  {;; main buttons
   :a         0x0001
   :b         0x0002

   ;; directional pad
   :r         0x0010
   :l         0x0020
   :u         0x0040
   :d         0x0080

   ;; meta buttons
   :select    0x0004
   :start     0x0008

   ;; pseudo-buttons
   :restart   0x0800 ; hard reset -- not really a button
   :listen -1        ; listen for user input
   })

(defn button-mask [buttons]
  (reduce bit-or 0x0000 (map button-code buttons)))

(def current-state (atom nil))

(defn step
  ([^SaveState state buttons]
     (if (not @on?) (restart!))
     (if (not= @current-state state)
       (Gb/loadState (:data state)))
     (Gb/step (button-mask buttons))
     (reset! current-state 
             (SaveState. (inc (:frame state))(Gb/saveState)))))

(defn play
  ([^SaveState state]
     (step state [:listen]))
  ([] (step (if @current-state @current-state (root)))))
  
